/*---------------------------------------------------------------------------------------------
 *  Copyright (c) Microsoft Corporation. All rights reserved.
 *  Licensed under the MIT License. See License.txt in the project root for license information.
 *--------------------------------------------------------------------------------------------*/
'use strict';
var vscode_languageserver_1 = require('vscode-languageserver');
var vscode_html_languageservice_1 = require('vscode-html-languageservice');
var languageModelCache_1 = require('./languageModelCache');
var nls = require('vscode-nls');
nls.config(process.env['VSCODE_NLS_CONFIG']);
// Create a connection for the server
var connection = vscode_languageserver_1.createConnection();
console.log = connection.console.log.bind(connection.console);
console.error = connection.console.error.bind(connection.console);
// Create a simple text document manager. The text document manager
// supports full document sync only
var documents = new vscode_languageserver_1.TextDocuments();
// Make the text document manager listen on the connection
// for open, change and close text document events
documents.listen(connection);
var htmlDocuments = languageModelCache_1.getLanguageModelCache(10, 60, function (document) { return vscode_html_languageservice_1.getLanguageService().parseHTMLDocument(document); });
documents.onDidClose(function (e) {
    htmlDocuments.onDocumentRemoved(e.document);
});
connection.onShutdown(function () {
    htmlDocuments.dispose();
});
var workspacePath;
// After the server has started the client sends an initilize request. The server receives
// in the passed params the rootPath of the workspace plus the client capabilites
connection.onInitialize(function (params) {
    workspacePath = params.rootPath;
    return {
        capabilities: {
            // Tell the client that the server works in FULL text document sync mode
            textDocumentSync: documents.syncKind,
            completionProvider: { resolveProvider: false, triggerCharacters: ['.', ':', '<', '"', '=', '/'] },
            documentHighlightProvider: true,
            documentRangeFormattingProvider: true,
            documentFormattingProvider: true,
            documentLinkProvider: true
        }
    };
});
// create the JSON language service
var languageService = vscode_html_languageservice_1.getLanguageService();
var languageSettings;
// The settings have changed. Is send on server activation as well.
connection.onDidChangeConfiguration(function (change) {
    var settings = change.settings;
    languageSettings = settings.html;
});
connection.onCompletion(function (textDocumentPosition) {
    var document = documents.get(textDocumentPosition.textDocument.uri);
    var htmlDocument = htmlDocuments.get(document);
    var options = languageSettings && languageSettings.suggest;
    return languageService.doComplete(document, textDocumentPosition.position, htmlDocument, options);
});
connection.onDocumentHighlight(function (documentHighlightParams) {
    var document = documents.get(documentHighlightParams.textDocument.uri);
    var htmlDocument = htmlDocuments.get(document);
    return languageService.findDocumentHighlights(document, documentHighlightParams.position, htmlDocument);
});
function merge(src, dst) {
    for (var key in src) {
        if (src.hasOwnProperty(key)) {
            dst[key] = src[key];
        }
    }
    return dst;
}
function getFormattingOptions(formatParams) {
    var formatSettings = languageSettings && languageSettings.format;
    if (!formatSettings) {
        return formatParams;
    }
    return merge(formatParams, merge(formatSettings, {}));
}
connection.onDocumentFormatting(function (formatParams) {
    var document = documents.get(formatParams.textDocument.uri);
    return languageService.format(document, null, getFormattingOptions(formatParams));
});
connection.onDocumentRangeFormatting(function (formatParams) {
    var document = documents.get(formatParams.textDocument.uri);
    return languageService.format(document, formatParams.range, getFormattingOptions(formatParams));
});
connection.onDocumentLinks(function (documentLinkParam) {
    var document = documents.get(documentLinkParam.textDocument.uri);
    return languageService.findDocumentLinks(document, workspacePath);
});
// Listen on the connection
connection.listen();
//# sourceMappingURL=https://ticino.blob.core.windows.net/sourcemaps/adb9c231677a914df9338475950832a9f1f7ffc5/extensions/html/server/out/htmlServerMain.js.map
